package edu.sc.seis.cormorant;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.omg.CORBA.ORB;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.POAHelper;
import edu.iris.Fissures.model.AllVTFactory;
import edu.sc.seis.cormorant.nameService.NameServiceController;
import edu.sc.seis.fissuresUtil.exceptionHandler.MailExceptionReporter;
import edu.sc.seis.fissuresUtil.namingService.FissuresNamingService;
import edu.sc.seis.fissuresUtil.simple.Initializer;

public class Start {

    public void init(String[] args) throws Exception {
        try {
            props = Initializer.loadProperties(args);
            if ( ! props.containsKey("jacorb.implname")) {
                props.put("jacorb.implname", "cormorant");
            }
            // configure logging from properties...
            PropertyConfigurator.configure(props);
            System.out.println("cormorant logging configured.");
            logger.debug("cormorant init()");
            //configure exception mailer
            MailExceptionReporter.addMailExceptionReporter(props);
            // Initialize the ORB.
            orb = (org.omg.CORBA_2_3.ORB)org.omg.CORBA.ORB.init(args, props);
            // register valuetype factories
            new AllVTFactory().register(orb);
            fisName = Initializer.createNamingService(orb, props);
            if(props.containsKey(CORMORANT_NAME_SERVER)
                    && props.getProperty(CORMORANT_NAME_SERVER).equals("true")) {
                logger.info("Starting the cormorant name server");
                NameServiceController b = new NameServiceController(props,
                                                                    "nameServer",
                                                                    orb);
                b.start(orb, fisName);
            }
            String addNS = props.getProperty(FissuresNamingService.ADDITIONAL_CORBALOC_PROP);
            if(addNS != null) {
                fisName.addOtherNameServiceCorbaLoc(addNS);
            }
            String names = props.getProperty(CORMORANT_SERVERS);
            if(names == null || names.length() == 0) {
                throw new Exception("Property " + CORMORANT_SERVERS
                        + " must not be empty.");
            }
            StringTokenizer nameTokenizer = new StringTokenizer(names, ",");
            while(nameTokenizer.hasMoreTokens()) {
                String name = nameTokenizer.nextToken();
                logger.info("Creating " + name + " server");
                String starterClassName = props.getProperty(CORMORANT_SERVERS
                        + "." + name + ".controller");
                Class starterClass = Class.forName(starterClassName);
                try {
                    Constructor cs = starterClass.getConstructor(new Class[] {Properties.class,
                                                                              String.class,
                                                                              ORB.class});
                    Controller b = (Controller)cs.newInstance(new Object[] {props,
                                                                            name,
                                                                            orb});
                    controllers.add(b);
                } catch(NoSuchMethodException e) {
                    logger.error("Unable to find constructor "
                            + starterClassName
                            + "(Properties confProps, String serverPropName, ORB orb). Other servers will be started.");
                }
            }
            if(controllers.size() == 0) {
                logger.error("No servers to start, quiting...");
                throw new Exception("No servers to start.");
            }
        } catch(Exception e) {
            logger.fatal("In init", e);
            throw e;
        } catch(Throwable e) {
            logger.fatal("In init", e);
            throw new Exception(e);
        }
    }

    public void start() throws Exception {
        logger.debug("cormorant start()");
        try {
            // Resolve Root POA
            POA rootPOA = POAHelper.narrow(orb.resolve_initial_references("RootPOA"));
            // Get a reference to the POA manager
            Thread t = new Thread() {

                public void run() {
                    logger.info("Running the ORB");
                    orb.run();
                    logger.info("ORB.run() done");
                }
            };
            org.omg.PortableServer.POAManager manager = rootPOA.the_POAManager();
            manager.activate();
            t.start();
            Iterator it = controllers.iterator();
            while(it.hasNext()) {
                Controller b = (Controller)it.next();
                b.start(orb, fisName);
            }
            logger.info("Cormorant started");
        } catch(Exception e) {
            logger.fatal("In start", e);
            throw e;
        } catch(Throwable e) {
            logger.fatal("In start", e);
            throw new Exception(e);
        }
    }

    public void stop() throws Exception {
        logger.debug("cormorant stop()");
        try {
            Iterator it = controllers.iterator();
            while(it.hasNext()) {
                Controller cur = (Controller)it.next();
                try {
                    cur.stop(orb, fisName);
                    logger.debug("Successfuly stopped " + cur.getServerPropName());
                } catch(Exception e) {
                    logger.error("Unable to stop "
                            + cur.getServerPropName(), e);
                    // who cares
                }
            }
            if(orb != null) {
                POA rootPOA = POAHelper.narrow(orb.resolve_initial_references("RootPOA"));
                org.omg.PortableServer.POAManager manager = rootPOA.the_POAManager();
                manager.hold_requests(true);
            }
        } catch(Throwable e) {
            logger.fatal(e);
        }
        logger.debug("after cormorant stop()");
    }

    public void destroy() {
        logger.debug("cormorant destroy");
        if(orb != null) {
            logger.debug("before orb shutdown");
            orb.shutdown(false);
            logger.debug("after orb shutdown");
        }
        logger.info("cormorant done");
    }
    
    public static void main(final String[] args) {
        // just in case of failures prior to PropertyConfigurator
        BasicConfigurator.configure();
        final Start cormorant = new Start();
        try {
            // Install signal handler for INT (aka control-c), to quit
            // gracefully
            DaemonSignalHandler.install("INT", new Runnable() {

                public void run() {
                    try {
                        logger.info("Stopping due to signal: INT");
                        cormorant.stop();
                        cormorant.destroy();
                    } catch(Exception e) {
                        logger.error(e);
                    }
                }
            });
            // Install signal handler for TERM, to quit gracefully
            DaemonSignalHandler.install("TERM", new Runnable() {

                public void run() {
                    try {
                        logger.info("Stopping due to signal: TERM");
                        cormorant.stop();
                        cormorant.destroy();
                    } catch(Exception e) {
                        logger.error(e);
                    }
                }
            });
            // Install signal handler for ALRM, do a stop start
            String whichOS = System.getProperty("os.name");
            if(whichOS.toLowerCase().indexOf("windows") == -1) {
                DaemonSignalHandler.install("HUP", new Runnable() {

                    public void run() {
                        try {
                            logger.info("Restarting due to signal: HUP");
                            cormorant.stop();
                            cormorant.start();
                        } catch(Exception e) {
                            logger.error(e);
                        }
                    }
                });
            } else {
                logger.info("Not installing a signal listener on ALRM since Windows doesn't support it.");
            }
            cormorant.init(args);
            cormorant.start();
        } catch(Exception e) {
            logger.fatal("startup exception in main", e);
            try {
                cormorant.stop();
                cormorant.destroy();
            } catch(Exception ee) {
                // oh well...
            }
        }
    }

    public static final String CORMORANT_NAME_SERVER = "cormorant.servers.nameServer";

    public static final String CORMORANT_SERVERS = "cormorant.servers";

    protected org.omg.CORBA_2_3.ORB orb = null;

    protected FissuresNamingService fisName = null;

    protected Properties props = null;

    protected List controllers = new ArrayList();
    
    private static final Logger logger = Logger.getLogger(Start.class);
} // PlottableStart
